/*******************************************************************************
* Licensed Materials - Property of HCL
* (c) Copyright HCL Technologies Ltd. 2017.  All Rights Reserved.
*
* Note to U.S. Government Users Restricted Rights:
* Use, duplication or disclosure restricted by GSA ADP Schedule
* Contract with IBM Corp.
*******************************************************************************/

/**
 * Deploy the Design Room Live server.
 * - start Mongo DB (unless it's already running)
 * - start the Design Room Live server using PM2
 *
 * @author Mattias Mohlin
 */

'use strict';

let path = require('path');
let conf = require('nconf');
require('nconf-strip-json-comments')(conf);
let mongo = require('mongodb');
let fs = require('fs');
let shell = require('shelljs');
let pm2 = require('pm2');
let term = require('terminal-kit').terminal;
let tmp = require('tmp');

conf.file('default_config', {
    file: './dr-config.json'
});

// If the DR server is running it will be stopped.
function stopDRServerIfRunning(callback) {
    pm2.describe('DesignRoomLive', (err, proc) => {
        if (err || proc.length === 0 || proc[0].pm2_env.status == 'stopped') {
            console.log('Design Room Live server is not running.');
            callback();
            return;
        }

        console.log('Design Room Live server is running. It will now be stopped!');
        // Stop by deleting the app from PM2 (in case a new version of the app is about to be deployed)
        pm2.delete('DesignRoomLive', (err) => {
            if (err) {
                console.log('Failed to stop Design Room Live server. Make sure PM2 is installed by running install-dr first.');
                console.log(err);
                process.exit(1);
            }

            console.log('Successfully stopped Design Room Live server!');
            callback();
        });
    });
}

// Start DesignRoom Live server in a mode where it just reports the configuration settings it will use.
// The configuration settings are provided as the first argument of the callback function.
function getDRServerConfig(pm2ConfigFile, callback) {
    // Connect to an existing PM2 daemon or launch a new one
    pm2.connect((err) => {
        if (err) {
            console.log('Failed to connect to PM2: ' + err);
            console.log('Make sure PM2 is installed by running install-dr first.');
            process.exit(1);
        }

        // Stop the DR server if it's already running
        stopDRServerIfRunning(() => {
            console.log('Fetching configuration settings from Design Room Live server. Launching it using settings from ' + pm2ConfigFile);

            tmp.file((err, tmpFilePath, fd, cleanupCallback) => {
                if (err) {
                    console.log('Failed to create temporary file for storing Design Room Live server configuration.');
                    console.log(err);
                    process.exit(1);
                }

                process.env['DR_REPORT_CONFIG'] = tmpFilePath;
                pm2.start(pm2ConfigFile, { autorestart: false }, (err, proc) => {
                    if (err) {
                        console.log('Failed to fetch configuration settings from Design Room Live server.');
                        console.log(err);
                        process.exit(1);
                    }

                    delete process.env.DR_REPORT_CONFIG;

                    // Wait until the the configuration file is available
                    function waitUntilConfigReady(attempts) {
                        if (attempts >= conf.get('dr_server_max_startup_time')) {
                            console.log('DR server failed to start within ' + conf.get('dr_server_max_startup_time') + ' seconds!');
                            process.exit(1);
                        }
                        setTimeout(() => {
                            if (fs.existsSync(tmpFilePath)) {
                        //        console.log('readtmpfilepath : ' + fs.readFileSync(tmpFilePath,'utf-8'));
                        //        console.log('tmpFilePath : ' + tmpFilePath);
                                try {
                                    let serverConfig = JSON.parse(fs.readFileSync(tmpFilePath, 'utf-8'));
                        //            console.log('readtmpfilepath : ' + fs.readFileSync(tmpFilePath,'utf-8'));
                                    console.log('Successfully read Design Room Live server configuration!');

                                    // Now stop the Design Room Live server before we proceed
                                    stopDRServerIfRunning(() => {
                                        console.log('Stopped Design Room Live server after reading the server configuration!');
                                        callback(serverConfig);
                                    });                                    
                                }
                                catch (err) {
                                    // Some error occurred. Perhaps the file is currently being written. Schedule another attempt!
                                    console.log('Server configuration settings not yet available after ' + attempts + ' seconds!');
                                    waitUntilConfigReady(attempts + 1);
                                }
                            }
                        }, 1000);
                    }
                    waitUntilConfigReady(1);
                });
            });

        });
    });
}

// Start Design Room Live server
function startDRServer(pm2ConfigFile, serverConfig, callback) {
    // Connect to an existing PM2 daemon or launch a new one
    pm2.connect((err) => {
        if (err) {
            console.log('Failed to connect to PM2: ' + err);
            console.log('Make sure PM2 is installed by running install-dr first.');
            process.exit(1);
        }

        // Stop the DR server if it's already running
        stopDRServerIfRunning(() => {
            console.log('Starting Design Room Live server now using settings from ' + pm2ConfigFile);
            pm2.start(pm2ConfigFile, (err, proc) => {
                if (err) {
                    console.log('Failed to start Design Room Live server.');
                    console.log(err);
                    process.exit(1);
                }

                console.log('Successfully started Design Room Live server!');
                console.log('Design Room Live server status is now: ' + proc[0].pm2_env.status);

                console.log('Server URL is: https://' + serverConfig.dr_host + ':' + serverConfig.dr_port + '/index.html');

                callback();
            });
        });
    });
}

// Attempt to connect to Mongo
function startMongoIfNeeded(mongoURL, mongoExe, dataFolder, port, logfile, callback) {
    let client = mongo.MongoClient;
    client.connect(mongoURL, function (err, db) {
        if (err) {
            console.log('Could not detect a running Mongo database at ' + mongoURL + '. Start it now...');

            if (!fs.existsSync(mongoExe)) {
                console.log('Failed to find ' + mongoExe + '! Did you forget to run install-dr?');
                process.exit(1);
            }

            if (!fs.existsSync(dataFolder)) {
                // The data folder must exist before we can start Mongo. Create it now.
                console.log('Folder ' + dataFolder + ' is missing. Creating it now.');
                shell.mkdir('-p', dataFolder);
                if (shell.error()) {
                    console.log('Failed to create folder ' + dataFolder);
                    process.exit(1);
                }
            }

            let out = fs.openSync(logfile, 'a');
            let err = fs.openSync(logfile, 'a');

            let spawn = require('child_process').spawn;

            let childProcess = spawn(mongoExe, ['--dbpath', dataFolder, '--port', port], { detached: true, stdio: ['ignore', out, err] });
            childProcess.unref();

            childProcess.on('error', function (err) {
                console.log('Failed to start Mongo!');
                process.exit(1);
            });

            childProcess.on('close', function (code) {
                console.log('Mongo terminated unexpectedly!');
                process.exit(1);
            });

            // Wait until Mongo is up and running            
            function waitUntilStarted(attempts) {
                if (attempts >= conf.get('mongo_max_startup_time')) {
                    console.log('Mongo failed to start within ' + conf.get('mongo_max_startup_time') + ' seconds!');
                    process.exit(1);
                }
                setTimeout(() => {
                    client.connect(mongoURL, function (err, db) {
                        if (err) {
                            console.log('Waited ' + attempts + ' of ' + conf.get('mongo_max_startup_time') + ' seconds...');
                            waitUntilStarted(attempts + 1);
                            return;
                        }
                        console.log('Detected a running Mongo database after ' + attempts + ' seconds!');
                        console.log('See mongostartup.log for details.');
                        callback();
                    });
                }, 1000);
            }
            waitUntilStarted(1);
        }
        else {
            console.log('Detected a running Mongo database, so it will not be started again.');
            callback();
        }
    });
}

if (process.platform == 'win32') {
    var mongoExe = conf.get('mongo_win_install_folder') + 'bin\\mongod.exe';
    var dataFolder = conf.get('mongo_win_data_folder');
}
else if (process.platform == 'linux') {
    var mongoExe = conf.get('mongo_linux_install_folder') + 'bin/mongod';
    var dataFolder = conf.get('mongo_linux_data_folder');
}

let mongoLogfile = conf.get('mongo_log_file');
let pm2ConfigFile = conf.get('pm2_config_file');

// As the very first step, launch the Design Room Live server in a mode where it just reports which configuration settings it will use.
// We need to know these real settings, rather than what is obtained from ../OnPrem_Design_Room/conf since the PM2 config file
// may override some of the settings.
getDRServerConfig(pm2ConfigFile, (serverConfig) => {

    const mongoURL = 'mongodb://' + serverConfig.dr_dbhost + ':' + serverConfig.dr_dbport + '/DesignRoomLive';

    // Start Mongo (if needed)
    startMongoIfNeeded(mongoURL, mongoExe, dataFolder, serverConfig.dr_dbport, mongoLogfile, () => {

        // Start Design Room Live server using PM2
        startDRServer(pm2ConfigFile, serverConfig, () => {
            console.log('All done!');
            process.exit(0);
        });

    });

});